package au.org.vicnode.mflux.services;

import java.util.List;

import arc.mf.plugin.PluginService;
import arc.mf.plugin.dtype.BooleanType;
import arc.mf.plugin.dtype.LongType;
import arc.mf.plugin.dtype.StringType;
import arc.xml.XmlDoc;
import arc.xml.XmlDoc.Element;
import arc.xml.XmlDocMaker;
import arc.xml.XmlWriter;
import au.org.vicnode.mflux.plugin.util.Project;
import au.org.vicnode.mflux.plugin.util.Properties;

public class SvcProjectRcpAllocationIdSet extends PluginService {

	public static final String SERVICE_NAME = Properties.SERVICE_ROOT + "project.rcp.allocation.id.set";

	private final Interface _defn;

	public SvcProjectRcpAllocationIdSet() {
		_defn = new Interface();
		_defn.add(new Interface.Element("project-id", StringType.DEFAULT, "The project id.", 1, 1));
		_defn.add(new Interface.Element("rcp-allocation-id", LongType.POSITIVE, "The RCP allocation id.", 1, 1));
		_defn.add(new Interface.Element("overwrite", BooleanType.DEFAULT,
				"Setting to true to overwrite the RCP allocation id if it pre-exists. Defaults to false.", 0, 1));
	}

	@Override
	public Access access() {
		return ACCESS_ADMINISTER;
	}

	@Override
	public Interface definition() {
		return _defn;
	}

	@Override
	public String description() {
		return "Set the rcp-allocation-id in the project's meta-data.";
	}

	@Override
	public String name() {
		return SERVICE_NAME;
	}

	@Override
	public void execute(Element args, Inputs inputs, Outputs outputs, XmlWriter w) throws Throwable {

		String projectId = args.value("project-id");
		String rcpAllocationId = args.value("rcp-allocation-id");
		boolean overwrite = args.booleanValue("overwrite", false);
		String projectNamespace = Project.assetNameSpace(executor(), projectId, false);
		String docType = Properties.getCollectionDocType(executor());

		// get current namespace/asset-meta
		XmlDoc.Element ne = executor().execute("asset.namespace.describe",
				"<args><namespace>" + projectNamespace + "</namespace></args>", null, null).element("namespace");
		// inherit-asset-meta is true if it is not present in the result of
		// asset.namespace.describe (in MF 4.17.023). This is confusing...
		boolean inheritAssetMeta = ne.booleanValue("inherit-asset-meta", true);
		XmlDoc.Element assetMeta = ne.element("/asset-meta");
		List<XmlDoc.Element> des = assetMeta.elements();
		if (assetMeta == null || des == null || !assetMeta.elementExists(docType)) {
			throw new Exception("Project " + projectId + " does not have asset-meta/" + docType);
		}

		String currentRcpAllocationId = assetMeta.value(docType + "/rcp-allocation-id");
		boolean exists = currentRcpAllocationId != null;
		boolean matches = rcpAllocationId.equals(currentRcpAllocationId);
		if (exists) {
			if (matches) {
				// already set and matches the value, no need to change
				w.add("updated", new String[] { "project-id", projectId, "rcp-allocation-id", rcpAllocationId, "exists",
						Boolean.toString(exists), "matches", Boolean.toString(matches) }, false);
				return;
			} else {
				if (!overwrite) {
					// already set but does not match the value, however, not overwrite
					w.add("updated",
							new String[] { "project-id", projectId, "rcp-allocation-id", currentRcpAllocationId,
									"new-rcp-allocation-id", rcpAllocationId, "exists", Boolean.toString(exists),
									"matches", Boolean.toString(matches) },
							false);
					return;
				}
			}
		}

		XmlDocMaker dm = new XmlDocMaker("args");
		dm.add("namespace", projectNamespace);
        dm.add("inherit-asset-meta", inheritAssetMeta);
		dm.push("asset-meta");
		for (XmlDoc.Element de : des) {
			if (docType.equals(de.qname())) {
				if (de.elementExists("rcp-allocation-id")) {
					de.element("rcp-allocation-id").setValue(rcpAllocationId);
				} else {
					de.add(new XmlDoc.Element("rcp-allocation-id", rcpAllocationId));
				}
			}
			dm.add(de, true);
		}
		dm.pop();
		executor().execute("asset.namespace.asset.meta.set", dm.root());

		if (exists) {
			w.add("updated", new String[] { "project-id", projectId, "rcp-allocation-id", rcpAllocationId, "exists",
					Boolean.toString(exists), "matches", Boolean.toString(matches) }, true);
		} else {
			w.add("updated", new String[] { "project-id", projectId, "rcp-allocation-id", rcpAllocationId, }, true);
		}

	}

}
